<?php

namespace Drupal\xnttbrapi\Plugin\ExternalEntities\StorageClient;

use Drupal\Core\Form\FormStateInterface;
use Drupal\external_entities\Plugin\ExternalEntities\StorageClient\Rest;
use GuzzleHttp\Exception\ClientException;
use GuzzleHttp\Exception\ConnectException;
use GuzzleHttp\Exception\RequestException;
use GuzzleHttp\Exception\ServerException;

/**
 * External entities storage client based on a REST API.
 *
 * @ExternalEntityStorageClient(
 *   id = "brapi",
 *   label = @Translation("BrAPI"),
 *   description = @Translation("Retrieves external entities from a Breeding API enpoint.")
 * )
 */
class Brapi extends Rest {

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(
    array $form,
    FormStateInterface $form_state
  ) {
    $form = parent::buildConfigurationForm($form, $form_state);

    // Add help.
    $form['endpoint']['#description'] = $this->t('Enter the BrAPI enpoint URL including curly brackets with object identifier. Ex.:<ul><li>"https://www.crop-diversity.org/mgis/brapi/v1/germplasm/{germplasmDbId}"</li><li>"https://www.crop-diversity.org/mgis/brapi/v1/germplasm?accessionNumber={accessionNumber}".</li></ul>Don\'t forget to map the external entity identifier to the field named inside the curly brackets.');
    $form['api_key']['key']['#description'] = $this->t('Enter here the BrAPI token to use if the endpoint requires one.');
    $form['api_key']['key']['#title'] = $this->t('BrAPI Token');
    // Remove 'Bearer ' part automatically added by form validation.
    $form['api_key']['key']['#default_value'] = str_replace(
      'Bearer ',
      '',
      $this->configuration['api_key']['key']
    );

    // Auto-set some REST parameters.
    $form['response_format'] = [
      '#type' => 'hidden',
      '#value' => 'json',
    ];

    $form['pager']['page_parameter'] = [
      '#type' => 'hidden',
      '#value' => 'page',
    ];

    $form['pager']['page_parameter_type'] = [
      '#type' => 'hidden',
      '#value' => 'pagenum',
    ];

    $form['pager']['page_size_parameter'] = [
      '#type' => 'hidden',
      '#value' => 'pageSize',
    ];

    $form['pager']['page_size_parameter_type'] = [
      '#type' => 'hidden',
      '#value' => 'pagesize',
    ];

    $form['api_key']['header_name'] = [
      '#type' => 'hidden',
      '#value' => 'Authorization',
    ];

    $form['parameters']['#type'] = 'details';
    $form['parameters']['#title'] = $this->t('Optional parameters');
    $form['parameters']['#open'] = FALSE;
    $form['parameters']['#description'] = $this->t('Optional query parameters can be specified to pre-filter results. For instance, use "instituteCode|TheInstituteName" to restrict germplasm to the institute "TheInstituteName" or "species|SomeSpeciesName" to restrict to the species "SpeciesName".');

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateConfigurationForm(
    array &$form,
    FormStateInterface $form_state
  ) {
    // Prefix key if needed.
    $api_key = $form_state->getValue('api_key');
    if (!empty($api_key['key'])
        && (@substr_compare($api_key['key'], 'Bearer', 0, 6) !== 0)
    ) {
      $api_key['key'] = 'Bearer ' . $api_key['key'];
      $form_state->setValue('api_key', $api_key);
    }

    parent::validateConfigurationForm($form, $form_state);
  }

  /**
   * Returns the BrAPI call URL.
   *
   * @param string $id
   *   A BrAPI entity identifier or an empty string if none for listing.
   *
   * @return string
   *   The BrAPI endpoint URL qualified with the given identifier if one.
   */
  public function getEndpoint(string $id = '') :string {
    $brapi_url = $this->configuration['endpoint'];
    // Add the entity identifier to the URL.
    if (preg_match('#\{\w+\}#', $brapi_url)) {
      // Use specified placeholder if one.
      // @todo: Maybe manage multiple placeholders and identifiers, for instance
      // "maps/{mapDbId}/positions/{linkageGroupName}" with "42/806"?
      if (empty($id)) {
        $brapi_url = preg_replace('#(?:[^/?&=;]+=|/)?\{\w+\}.*#', '', $brapi_url);
      }
      else {
        $brapi_url = preg_replace('#\{\w+\}#', $id, $brapi_url);
      }
    }
    else {
      // Append otherwise.
      $brapi_url = rtrim($brapi_url, '/') . '/' . $id;
    }

    return $brapi_url;
  }

  /**
   * {@inheritdoc}
   */
  public function load($id) {

    $brapi_url = $this->getEndpoint($id);
    // Support query string in provided BrAPI endpoint.
    list($brapi_url, $brapi_query) = explode('?', $brapi_url) + [1 => ''];
    if (!empty($brapi_query)) {
      // Note: we don't consider ';' as a query parameter separator as it is not
      // used by BrAPI specifications.
      preg_match_all('#([^&=]+)(?:=([^&=]*))?#', $brapi_query, $matches);
      $brapi_query = array_combine($matches[1], $matches[2]);
    }
    else {
      $brapi_query = [];
    }
    // Endpoint parameters come first (object ID won't be overriden).
    $query = $brapi_query + $this->getSingleQueryParameters($id);
    try {
      $response = $this->httpClient->request(
        'GET',
        $brapi_url,
        [
          'headers' => $this->getHttpHeaders(),
          'query' => $query,
        ]
      );

      $format = $this->configuration['response_format'];
      // Remove BOM.
      $body = ltrim($response->getBody(), "\xEF\xBB\xBF");

      $result = $this
        ->getResponseDecoderFactory()
        ->getDecoder($format)
        ->decode($body)
      ;
    }
    catch (ServerException $e) {
      \Drupal::messenger()->addWarning(
        $this->t('Failed to query BrAPI endpoint. See logs for details.')
      );
      $response = $e->getResponse();
      $response_body = $response->getBody()->getContents();
      \Drupal::logger('xnttbrapi')->warning(
        'Failed to query BrAPI endpoint: '
        . $response_body
      );
    }
    catch (ClientException $e) {
      // Not found or some other 4xx error.
    }
    catch (ConnectException | RequestException $e) {
      \Drupal::messenger()->addWarning(
        $this->t('Failed to connect to BrAPI endpoint. See logs for details.')
      );
      \Drupal::logger('xnttbrapi')->warning(
        'Failed to connect to BrAPI endpoint: '
        . $e
      );
    }
    // To work with search calls or GET filters, we need to support one object
    // returned into the 'data' key.
    if (isset($result['result']['data'])) {
      if (!empty($result['result']['data'])) {
        $data = $result['result']['data'][0];
        // More than one object means the filter did not work to return a single
        // object as expected.
        if (1 < count($result['result']['data'])) {
          \Drupal::logger('xnttbrapi')->warning(
            'The given endpoint returned more than one object for ID "'
            . $id
            . '" from: '
            . $this->getEndpoint($id)
          );
        }
      }
      else {
        $data = NULL;
      }
    }
    else {
      // Otherwise, regular call result: the object is into the "result" key.
      $data = $result['result'] ?? NULL;
    }
    return $data;
  }

  /**
   * {@inheritdoc}
   */
  public function query(
    array $parameters = [],
    array $sorts = [],
    $start = NULL,
    $length = NULL
  ) {
    // Note: the default $length to 50 is set by ExternalEntityListBuilder (hard
    // coded) so the default number of item per page is not taken into account
    // on regular external entity listing pages.
    // See https://www.drupal.org/project/external_entities/issues/3136353.

    $brapi_url = $this->getEndpoint();
    // Support query string in provided BrAPI endpoint.
    list($brapi_url, $brapi_query) = explode('?', $brapi_url) + [1 => ''];
    if (!empty($brapi_query)) {
      // Note: we don't consider ';' as a query parameter separator as it is not
      // used by BrAPI specifications.
      preg_match_all('#([^&=]+)(?:=([^&=]*))?#', $brapi_query, $matches);
      $brapi_query = array_combine($matches[1], $matches[2]);
    }
    else {
      $brapi_query = [];
    }
    // Pager and list parameters come first (discard endpoint parameter
    // overrides for listing).
    $query =
      $this->getListQueryParameters($parameters, $start, $length)
      + $brapi_query
    ;
    try {
      $response = $this->httpClient->request(
        'GET',
        $this->getEndpoint(),
        [
          'headers' => $this->getHttpHeaders(),
          'query' => $query,
        ]
      );

      $format = $this->configuration['response_format'];
      $body = ltrim($response->getBody(), "\xEF\xBB\xBF");

      $results = $this
        ->getResponseDecoderFactory()
        ->getDecoder($format)
        ->decode($body);
    }
    catch (ServerException | ClientException $e) {
      \Drupal::messenger()->addWarning(
        $this->t('Failed to query BrAPI endpoint. See logs for details.')
      );
      $response = $e->getResponse();
      $response_body = $response->getBody()->getContents();
      \Drupal::logger('xnttbrapi')->warning(
        'Failed to query BrAPI endpoint: '
        . $response_body
      );
    }
    catch (ConnectException | RequestException $e) {
      \Drupal::messenger()->addWarning(
        $this->t('Failed to connect to BrAPI endpoint. See logs for details.')
      );
      \Drupal::logger('xnttbrapi')->warning(
        'Failed to connect to BrAPI endpoint: '
        . $e
      );
    }

    $items = [];
    if (!empty($results['result']['data'])) {
      $items = array_values($results['result']['data']);
    }

    return $items;
  }

  /**
   * {@inheritdoc}
   */
  public function countQuery(array $parameters = []) {
    $brapi_url = $this->getEndpoint();
    // Support query string in provided BrAPI endpoint.
    list($brapi_url, $brapi_query) = explode('?', $brapi_url) + [1 => ''];
    if (!empty($brapi_query)) {
      // Note: we don't consider ';' as a query parameter separator as it is not
      // used by BrAPI specifications.
      preg_match_all('#([^&=]+)(?:=([^&=]*))?#', $brapi_query, $matches);
      $brapi_query = array_combine($matches[1], $matches[2]);
    }
    else {
      $brapi_query = [];
    }
    // Pager and list parameters come first (discard endpoint parameter
    // overrides for listing).
    $query = $this->getListQueryParameters($parameters, 0, 1) + $brapi_query;
    try {
      $response = $this->httpClient->request(
        'GET',
        $brapi_url,
        [
          'headers' => $this->getHttpHeaders(),
          'query' => $query,
        ]
      );

      $format = $this->configuration['response_format'];
      $body = ltrim($response->getBody(), "\xEF\xBB\xBF");

      $results = $this
        ->getResponseDecoderFactory()
        ->getDecoder($format)
        ->decode($body);
    }
    catch (ServerException | ClientException $e) {
      \Drupal::messenger()->addWarning(
        $this->t('Failed to count BrAPI endpoint. See logs for details.')
      );
      $response = $e->getResponse();
      $response_body = $response->getBody()->getContents();
      \Drupal::logger('xnttbrapi')->warning(
        'Failed to count BrAPI endpoint: '
        . $response_body
      );
    }
    catch (ConnectException | RequestException $e) {
      \Drupal::messenger()->addWarning(
        $this->t('Failed to connect to BrAPI endpoint. See logs for details.')
      );
      \Drupal::logger('xnttbrapi')->warning(
        'Failed to connect to BrAPI endpoint: '
        . $e
      );
    }

    $count = 0;
    // Tries to get data count from the API.
    if (!empty($results['metadata']['pagination']['totalCount'])) {
      $count = $results['metadata']['pagination']['totalCount'];
    }
    else {
      // Fallback to default if it failed.
      $count = parent::countQuery($parameters);
    }

    return $count;
  }
}
